from .python import PythonBuilder
from ..outputs.python import PythonOutput


class SetupPyBuilder(PythonBuilder):
    TYPE = "setup_py"

    def finish(self):
        super().finish()

        python_output: PythonOutput = self.builder.get_output("python", "setup.py")
        python_output.merge("setup_py", {"settings": self.settings})



from pathlib import Path

from . import process, OutputBuilder
from .json_base import JSONBaseBuilder
from .utils import ensure_parent_modules
from ..outputs.cfg import CFGOutput


class InheritedModelBuilder(JSONBaseBuilder):
    TYPE = "inherited_model"
    output_file_type = "json"
    output_file_name = "inherited-model-file"
    parent_module_root_name = "models"

    @process("**")
    def model_element(self):
        self.model_element_enter()
        self.build_children()
        self.model_element_leave()

    def enter_model(self):
        # remove the special json base builder functionality
        # TODO: better handling for this
        pass

    def output_primitive(self, top, data):
        if data is not None:
            if not isinstance(data, (str, float, int, bool)):
                data = str(data)
        if top.key == "for-inheritance":
            return
        return super().output_primitive(top, data)

    def begin(self, schema, settings):
        super().begin(schema, settings)

        output_name = self.settings[self.output_file_name]
        self.output = self.builder.get_output(self.output_file_type, output_name)

        ensure_parent_modules(
            self.builder, Path(output_name), ends_at=self.parent_module_root_name
        )

    def finish(self):
        # force clean output
        self.output.force_clean_output()



from oarepo_model_builder.builders.inherited_model import InheritedModelBuilder
from oarepo_model_builder.invenio.invenio_record_schema import (
    InvenioRecordSchemaBuilder,
)
from oarepo_model_builder.property_preprocessors import PropertyPreprocessor, process
from oarepo_model_builder.stack import ModelBuilderStack
from oarepo_model_builder.utils.camelcase import camel_case


class MarshmallowClassGeneratorPreprocessor(PropertyPreprocessor):
    TYPE = "marshmallow_class_generator"

    @process(
        model_builder=InvenioRecordSchemaBuilder,
        path="**/properties",
        condition=lambda current, stack: stack.schema_valid,
    )
    def modify_object_marshmallow(self, data, stack: ModelBuilderStack, **kwargs):
        schema_element_type = stack.top.schema_element_type

        if schema_element_type == "properties":
            for k, v in stack.top.data.items():
                self.add_class_name(stack, k, v)
        elif schema_element_type == "items":
            self.add_class_name(stack, None, stack.top.data)
        return data

    @process(model_builder=InvenioRecordSchemaBuilder, path="/model$")
    def modify_object_marshmallow_model(self, data, stack: ModelBuilderStack, **kwargs):
        self.add_root_class_name(stack)
        return data

    @process(
        model_builder=InheritedModelBuilder,
        path="**/properties",
        condition=lambda current, stack: stack.schema_valid,
        priority=1,
    )
    def modify_object_marshmallow_for_inherited(
        self, data, stack: ModelBuilderStack, **kwargs
    ):
        schema_element_type = stack.top.schema_element_type
        # TODO: if top-level, add to base classes, else add 'generate: False'
        if schema_element_type == "properties":
            for k, v in stack.top.data.items():
                definition = self.add_class_name(stack, k, v)
                if definition:
                    definition["generate"] = False
        elif schema_element_type == "items":
            definition = self.add_class_name(stack, None, stack.top.data)
            if definition:
                definition["generate"] = False
        return data

    @process(
        model_builder=InheritedModelBuilder,
        path="**/properties/**",
        condition=lambda current, stack: stack.schema_valid
        and stack.top.schema_element_type == "property",
        priority=1,
    )
    def modify_object_marshmallow_for_inherited(
        self, data, stack: ModelBuilderStack, **kwargs
    ):
        marshmallow = data.setdefault("oarepo:marshmallow", {})
        marshmallow["read"] = False
        marshmallow["write"] = False
        return data

    @process(model_builder=InheritedModelBuilder, path="/model$")
    def modify_object_marshmallow_model_for_inherited(
        self, data, stack: ModelBuilderStack, **kwargs
    ):
        definition = self.add_root_class_name(stack)
        definition["base-classes"] = [definition.pop("class")]
        return data

    @process(
        model_builder=InheritedModelBuilder,
        path="/model/properties/metadata$",
        priority=5,
    )
    def modify_object_marshmallow_metadata_for_inherited(
        self, data, stack: ModelBuilderStack, **kwargs
    ):
        definition = self.add_class_name(stack, "metadata", stack.top.data)
        definition["base-classes"] = [definition.pop("class")]
        definition["generate"] = True
        return data

    def add_class_name(self, stack, key, data):
        if "items" not in data and "properties" not in data:
            return
        definition = data.setdefault("oarepo:marshmallow", {})
        if "class" in definition:
            return definition
        definition["generate"] = True
        if key:
            class_name = key
        else:
            class_name = self.get_property_name(stack)
        class_name = (
            self.settings.python.record_prefix + camel_case(class_name) + "Schema"
        )
        definition["class"] = f"{self.settings.package}.services.schema." + class_name
        return definition

    def add_root_class_name(self, stack):
        data = stack.top.data
        definition = data.setdefault("oarepo:marshmallow", {})
        if "class" in definition:
            return definition
        definition["generate"] = True
        definition[
            "class"
        ] = f"{self.settings.package}.services.schema.{self.settings.python.record_prefix}RecordSchema"
        return definition

    def get_property_name(self, stack):
        for el in reversed(stack.stack):
            if el.schema_element_type == "property":
                return el.key
        return "UnknownName"



from oarepo_model_builder.invenio.invenio_record_schema import (
    InvenioRecordSchemaBuilder,
)
from oarepo_model_builder.property_preprocessors import PropertyPreprocessor, process
from oarepo_model_builder.stack import ModelBuilderStack
from oarepo_model_builder.utils.deepmerge import deepmerge


class ValidatorsPreprocessor(PropertyPreprocessor):
    @process(
        model_builder=InvenioRecordSchemaBuilder,
        path="**/properties/*",
    )
    def validators_marshmallow(self, data, stack: ModelBuilderStack, **kwargs):
        try:
            definition = data["oarepo:marshmallow"]
            validators = definition["validators"]
        except:
            validators = []
        try:
            minimum = data.get("minimum", None)
            maximum = data.get("maximum", None)
            exclusive_minimum = data.get("exclusiveMinimum", None)
            exclusive_maximum = data.get("exclusiveMaximum", None)

            if (
                minimum is not None
                or maximum is not None
                or exclusive_minimum is not None
                or exclusive_maximum is not None
            ):
                validators.append(
                    f"ma_valid.Range("
                    f'min={minimum or exclusive_minimum or "None"}, max={maximum or exclusive_maximum or "None"}, '
                    f"min_inclusive={exclusive_minimum is None}, max_inclusive={exclusive_maximum is None})"
                )
            min_length = data.get("minLength", None)
            max_length = data.get("maxLength", None)
            required = data.get("required", False)
            if required:
                deepmerge(
                    data.setdefault("oarepo:marshmallow", {}),
                    {"required": True},
                )
                print(type(data))
            if min_length is not None or max_length is not None:
                validators.append(
                    f"ma_valid.Length(min={min_length}, max={max_length})"
                )

            if len(validators) > 0:

                deepmerge(
                    data["oarepo:marshmallow"].setdefault("validators", []),
                    validators,
                )
        except:
            pass

        return data



from ..builders import OutputBuilder
from ..outputs.cfg import CFGOutput


class InvenioRecordResourceSetupCfgBuilder(OutputBuilder):
    TYPE = "invenio_record_resource_setup_cfg"

    def finish(self):
        super().finish()

        output: CFGOutput = self.builder.get_output("cfg", "setup.cfg")

        register_function = self.settings.python.create_blueprint_from_app.rsplit(
            ".", maxsplit=1
        )

        output.add_entry_point(
            "invenio_base.api_blueprints",
            self.settings.python.record_resource_blueprint_name,
            f"{register_function[0]}:{register_function[-1]}",
        )



import os
from pathlib import Path
from typing import Dict

from ..schema import ModelSchema
from . import ModelPreprocessor


class DefaultValuesModelPreprocessor(ModelPreprocessor):
    TYPE = "default"

    def transform(self, schema: ModelSchema, settings: Dict):
        self.set(
            settings,
            "package",
            lambda: os.path.basename(
                schema.schema.get("output-directory", os.getcwd())
            ).replace("-", "_"),
        )

        self.set(settings, "processing-order", lambda: ["settings", "*", "model"])

        self.set(
            settings,
            "package-base",
            lambda: settings.package.rsplit(".", maxsplit=1)[-1],
        )

        self.set(settings, "package-base-upper", lambda: settings.package_base.upper())

        self.set(settings, "kebap-package", lambda: settings.package.replace("_", "-"))

        @self.set(settings, "package-path")
        def c():
            package_path = settings.package.split(".")
            return Path(package_path[0]).joinpath(*package_path[1:])

        self.set(
            settings, "schema-version", lambda: schema.schema.get("version", "1.0.0")
        )

        self.set(
            settings,
            "schema-name",
            lambda: f"{settings.kebap_package}-{settings.schema_version}.json",
        )

        self.set(
            settings,
            "schema-file",
            lambda: os.path.join(
                settings.package_path, "records", "jsonschemas", settings.schema_name
            ),
        )

        self.set(
            settings, "mapping-package", lambda: f"{settings.package}.records.mappings"
        )

        self.set(
            settings,
            "jsonschemas-package",
            lambda: f"{settings.package}.records.jsonschemas",
        )

        self.set(
            settings,
            "mapping-file",
            lambda: os.path.join(
                settings.package_path,
                "records",
                "mappings",
                "v7",
                settings.package,
                settings.schema_name,
            ),
        )

        self.set(settings, "schema-server", lambda: "http://localhost/schemas/")

        self.set(
            settings,
            "index-name",
            lambda: settings.package
            + "-"
            + os.path.basename(settings.mapping_file).replace(".json", ""),
        )

        self.set(settings, "collection-url", lambda: f"/{settings.package_base}/")
        self.set(settings, "model-name", lambda: settings.package_base)

        # for outputting the model
        self.set(
            settings,
            "saved-model-file",
            lambda: settings.package_path / "models" / "model.json",
        )
        self.set(
            settings,
            "inherited-model-file",
            lambda: settings.package_path / "models" / "inherited_model.json",
        )



import copy
import importlib.resources
import json

import json5
import lazy_object_proxy
import pkg_resources
from jsonschema import Draft202012Validator
from jsonschema.exceptions import relevance

from oarepo_model_builder.utils.deepmerge import deepmerge


class InvalidModelException(Exception):
    pass


@lazy_object_proxy.Proxy
def model_json_schema():
    main_schema_path = (
        importlib.resources.files("oarepo_model_builder.validation")
        / "schemas"
        / "common-schema.json5"
    )
    schema = json5.loads(main_schema_path.read_text())
    for ep in pkg_resources.iter_entry_points(group="oarepo.model_schemas"):
        filename = ".".join(ep.attrs)
        data = pkg_resources.resource_string(ep.module_name, filename)
        data = json5.loads(data)
        schema["$defs"] = deepmerge(data, schema["$defs"], listmerge="extend")

    return schema


def validate_model(model, extra_validation_schemas=None):
    schema = copy.deepcopy(model_json_schema)
    if extra_validation_schemas:
        for e in extra_validation_schemas:
            schema["$defs"] = deepmerge(e, schema["$defs"], listmerge="extend")
    else:
        schema = model_json_schema

    data = json.loads(json.dumps(model.schema, default=lambda s: repr(s)))
    replace_array_keys(data)

    if "oarepo:model-validation" in data:
        schema["$defs"] = deepmerge(
            data.pop("oarepo:model-validation"), schema["$defs"], listmerge="extend"
        )

    validator = Draft202012Validator(schema)

    errors = list(validator.iter_errors(data))

    if not errors:
        return True

    errors.sort(key=lambda e: (relevance(e), e.path))
    print("\nErrors (most relevant first):")
    for err in errors:
        print(f'    on path "/{"/".join(str(x) for x in err.path)}" : {err.message}')
        print(f'         schema path "/{"/".join(str(x) for x in err.schema_path)}"')
    raise InvalidModelException("Invalid model")


def replace_array_keys(schema):
    if isinstance(schema, (list, tuple)):
        for l in schema:
            replace_array_keys(l)
    elif isinstance(schema, dict):
        for k, v in list(schema.items()):
            replace_array_keys(v)
            if k.endswith("[]"):
                del schema[k]
                schema[k[:-2]] = v



from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioRecordServiceConfigBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_record_service_config"
    class_config = "record-service-config-class"
    template = "record-service-config"



from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioScriptImportSampleDataBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_script_import_sample_data"
    class_config = "script-import-sample-data-cli"
    template = "script-import-sample-data"
    parent_modules = False

    def process_template(self, python_path, template, **extra_kwargs):
        super().process_template(python_path, template, **extra_kwargs)



from pathlib import Path

from ..utils.verbose import log
from . import process
from .json_base import JSONBaseBuilder
from .utils import ensure_parent_modules


class MappingBuilder(JSONBaseBuilder):
    TYPE = "mapping"
    output_file_type = "mapping"
    output_file_name = "mapping-file"
    parent_module_root_name = "mappings"

    @process("/model/**", condition=lambda current, stack: stack.schema_valid)
    def enter_model_element(self):
        # ignore schema leaves different than "type" - for example, minLength, ...
        # that should not be present in mapping
        element_type = self.stack.top.schema_element_type
        if element_type in ("properties", "property", "type", "items"):
            if element_type == "items":
                # do not output "items" container
                self.build_children()
                self.merge_mapping(self.stack.top.data)
                return
            elif element_type == "type" and self.stack.top.data == "array":
                # do not output "type=array"
                return

            self.model_element_enter()

            # process children
            self.build_children()

            self.merge_mapping(self.stack.top.data)

            self.model_element_leave()

    def merge_mapping(self, data):
        if isinstance(data, dict) and "oarepo:mapping" in data:
            mapping = self.call_components(
                "before_merge_mapping", data["oarepo:mapping"], stack=self.stack
            )
            self.output.merge_mapping(mapping)

    def on_enter_model(self, output_name):
        ensure_parent_modules(
            self.builder, Path(output_name), ends_at=self.parent_module_root_name
        )
        self.output.merge(
            self.settings.elasticsearch.templates[self.settings.elasticsearch.version]
        )
        self.output.enter("mappings", {})

    def finish(self):
        super().finish()

        log(
            log.INFO,
            f"""
    invenio index init --force
            """,
        )
